import asyncio
import json
import logging
from argparse import BooleanOptionalAction
from dataclasses import asdict
from typing import Any, Callable

import gi

from .metric import Gth

gi.require_version("Gio", "2.0")
from gi.repository import Gio, GLib  # type: ignore

GTH_UUID = "0000ec88-0000-1000-8000-00805f9b34fb"


log = logging.getLogger(__name__)


async def set_discovery(adapter_proxy, on=True) -> bool:
    method = "StartDiscovery" if on else "StopDiscovery"
    try:
        await adapter_proxy.call(method, None, Gio.DBusCallFlags.NO_AUTO_START, 500)  # type: ignore
        return True
    except GLib.Error as error:
        log.warning(f"{method}: {error.message}")
        return False


async def get_property(props_proxy, iface: str, property: str) -> Any:
    return (
        await props_proxy.call(  # type: ignore
            "Get",
            GLib.Variant("(ss)", (iface, property)),
            Gio.DBusCallFlags.NONE,
            -1,
        )
    ).unpack()[0]


class GthScanner:
    def __init__(self, alias_mapping: dict[str, str]):
        self.alias_mapping = alias_mapping
        self.queue = asyncio.Queue()

    async def scan_beacons(self, bluetooth_adapter_name: str = "hci0") -> asyncio.Queue:
        self._bluez_object_manager = Gio.DBusObjectManagerClient.new_for_bus_sync(
            Gio.BusType.SYSTEM, Gio.DBusObjectManagerClientFlags.DO_NOT_AUTO_START, "org.bluez", "/", None, None, None
        )
        self._bluez_object_manager.connect("object-added", self._object_added)
        adapters = [object for object in self._bluez_object_manager.get_objects() if object.get_interface("org.bluez.Adapter1")]
        matching_adapters = [adapter for adapter in adapters if adapter.get_object_path().endswith(bluetooth_adapter_name)]

        if not matching_adapters:
            raise RuntimeError(f"No bluez adapters for name {bluetooth_adapter_name} found!")
        adapter = matching_adapters[0]

        adapter_proxy = adapter.get_interface("org.bluez.Adapter1")
        adapter_props_proxy = adapter.get_interface("org.freedesktop.DBus.Properties")
        if not (adapter_proxy and adapter_props_proxy):
            raise RuntimeError("No usable bluez adapters found!")
        log.debug(f"Using adapter {adapter.get_object_path()}")
        try:
            await adapter_proxy.call(  # type: ignore
                "SetDiscoveryFilter",
                GLib.Variant("(a{sv})", ({"UUIDs": GLib.Variant("as", [GTH_UUID])},)),
                Gio.DBusCallFlags.NONE,
                -1,
            )
        except GLib.Error as err:
            raise RuntimeError(f"Failed to SetDiscoveryFilter: {err}")

        self._adapter_proxy = adapter_proxy
        self._adapter_props_proxy = adapter_props_proxy
        if await get_property(adapter_props_proxy, "org.bluez.Adapter1", "Discovering"):
            await set_discovery(adapter_proxy, False)
        await set_discovery(adapter_proxy, True)
        return self.queue  # FIXME, use closure

    def _object_added(self, __adapter__, dbus_object: Gio.DBusObject):
        if dbus_object.get_interface("org.bluez.Device1") and (props_proxy := dbus_object.get_interface("org.freedesktop.DBus.Properties")):

            async def handle_properties():
                address = await get_property(props_proxy, "org.bluez.Device1", "Address")
                alias = await get_property(props_proxy, "org.bluez.Device1", "Alias")
                if (new_alias := self.alias_mapping.get(address)) and (new_alias != alias):
                    alias = new_alias
                    await props_proxy.call(  # type: ignore
                        "Set",
                        GLib.Variant(
                            "(ssv)",
                            (  # type:ignore
                                "org.bluez.Device1",
                                "Alias",
                                GLib.Variant.new_string(new_alias),
                            ),
                        ),
                        Gio.DBusCallFlags.NONE,
                        -1,
                    )
                    print(f"Property set to {new_alias}")
                    uuids = await get_property(props_proxy, "org.bluez.Device1", "UUIDs")
                    manufacturer_data = await get_property(props_proxy, "org.bluez.Device1", "ManufacturerData")
                    rssi = await get_property(props_proxy, "org.bluez.Device1", "RSSI")
                    if (GTH_UUID in uuids) and (data := manufacturer_data.get(1)) and len(data) >= 6:
                        n = int.from_bytes(data[2:5], "big", signed=True)
                        temp = n // 1000 / 10
                        hum = n % 1000 / 10
                        batt = int(data[5] & 0x7F)
                        err = bool(data[5] & 0x80)
                        if not err:
                            self.queue.put_nowait(Gth(alias, address, rssi, temp, hum, batt))

            asyncio.create_task(handle_properties())



import argparse
import logging
import os
import sys
from dataclasses import dataclass
from time import sleep

import gi.events  # type: ignore

from .graphite import Graphite
from .prometheus import PushGateway

log = logging.getLogger(__name__)
import asyncio

import gi

gi.require_version("Gio", "2.0")
from gi.repository import GLib  # type: ignore

from gth_exporter.bluez import GthScanner
from gth_exporter.metric import Gth

log = logging.getLogger(__name__)


LOGLEVELS = {
    "DEBUG": logging.debug,
    "INFO": logging.info,
    "WARNING": logging.warning,
    "ERROR": logging.error,
    "CRITICAL": logging.critical,
}


def setup_logging(args):
    logging.basicConfig(
        level=args.log_level,
        format=("%(levelname)-8s %(message)s" if os.environ.get("JOURNAL_STREAM") else "%(asctime)s %(levelname)-8s %(message)s"),
    )


def main():
    parser = argparse.ArgumentParser(
        prog="GoveeLife Bluetooth Thermometer Hygrometer Exporter",
        description="Export metrics. Use environment variables METRICS_USER and METRICS_PASSWORD for authentication",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-l",
        "--log-level",
        choices=LOGLEVELS.keys(),
        help="log level",
        default="WARNING",
    )
    parser.add_argument(
        "-a",
        "--alias",
        default=[],
        type=str,
        action="append",
        help="Set aliases for specified device (e.g.: C4:7C:8D:XX:YY:ZZ=Bromeliad). Can be repeated",
    )
    parser.add_argument("-t", "--timeout", type=int, default=60, help="Scan timeout")
    parser.add_argument("-b", "--bluetooth-adapter", type=str, default="hci0", help="Bluetooth Adapter to use")
    parser.add_argument("-g", "--graphite-url", type=str, required=False, help="Post Metrics to Graphite Metrics URL")
    parser.add_argument("-p", "--prometheus-url", type=str, required=False, help="Post Metrics to Prometheus Pushgateway URL")
    args = parser.parse_args(sys.argv[1:])
    setup_logging(args)
    alias_mapping = dict([alias_s.split("=") for alias_s in args.alias])
    policy = gi.events.GLibEventLoopPolicy()
    mainloop = policy.get_event_loop()
    # Graphite Support
    if args.graphite_url:
        graphite = Graphite(args.graphite_url, os.getenv("METRICS_USER"), os.getenv("METRICS_PASSWORD"))
    else:
        graphite = None
    # Prometheus Pushgateway support
    if args.prometheus_url:
        prometheus = PushGateway(args.prometheus_url, os.getenv("METRICS_USER"), os.getenv("METRICS_PASSWORD"))
    else:
        prometheus = None

    gth_scanner = GthScanner(alias_mapping)  # FIXME: Await metrics_callback

    async def fun():
        try:
            async with asyncio.timeout(args.timeout):
                beacons = await gth_scanner.scan_beacons(args.bluetooth_adapter)
                while gth := await beacons.get():
                    tasks = []
                    if graphite:
                        tasks.append(asyncio.create_task(graphite.send_message(gth)))
                    if prometheus:
                        tasks.append(asyncio.create_task(prometheus.send_message(gth)))
                    await asyncio.wait(tasks)
                    print(gth)
        except TimeoutError:
            ...

    mainloop.run_until_complete(fun())  # FIXME Hack


if __name__ == "__main__":
    main()



import logging

import gi

gi.require_version("Soup", "3.0")

from gi.repository import Gio, GLib, Soup  # type: ignore

log = logging.getLogger(__name__)


from .metric import Gth


def to_prometheus(gth: Gth) -> str:
    return f"""
# TYPE govee_temperature_celsius gauge
govee_temperature_celsius{{alias="{gth.alias}", address="{gth.address}"}} {gth.temp_celsius}
# TYPE govee_humidity_percent gauge
govee_humidity_percent{{alias="{gth.alias}", address="{gth.address}"}} {gth.humidity_percent}
# TYPE govee_battery_percent gauge
govee_battery_percent{{alias="{gth.alias}", address="{gth.address}"}} {gth.battery_percent}
# TYPE govee_rssi_dbm gauge
# help Received Signal Strength Indicator
govee_rssi_dbm{{alias="{gth.alias}", address="{gth.address}"}} {gth.rssi}
"""


class PushGateway:
    url: str
    user: str | None
    password: str | None
    _session: Soup.Session

    def __init__(self, url: str, user: str | None = None, password: str | None = None):
        self.url = url
        self.user = user
        self.password = password
        self._session = Soup.Session()

        if log.getEffectiveLevel() == logging.DEBUG:
            logger = Soup.Logger.new(Soup.LoggerLogLevel.BODY)
            self._session.add_feature(logger)

    async def send_message(self, gth: Gth):
        uri = GLib.Uri.parse(self.url, GLib.UriFlags.NONE)
        message = Soup.Message.new_from_uri("POST", uri)
        assert self._session
        if self.user and self.password:
            auth_manager = self._session.get_feature(Soup.AuthManager)
            assert auth_manager
            auth = Soup.Auth.new(Soup.AuthBasic, message, "Basic")
            assert auth
            auth.authenticate(self.user, self.password)
            auth_manager.use_auth(message.get_uri(), auth)  # type: ignore

        body = to_prometheus(gth)

        message.set_request_body_from_bytes("application/x-www-form-urlencoded", GLib.Bytes.new(body.encode()))  # type: ignore

        resp_body = await self._session.send_and_read_async(message, GLib.PRIORITY_DEFAULT)  # type: ignore

        if (status := message.get_status()) != Soup.Status.OK:
            log.warning(f"Error Posting to '{self.url}': {Soup.Status.get_phrase(status)} -> {resp_body.get_data().decode()}")



import json
import logging
import socket
import time

import gi

gi.require_version("Soup", "3.0")

from gi.repository import Gio, GLib, Soup  # type: ignore

log = logging.getLogger(__name__)

from .metric import Gth


def to_graphite(gth: Gth) -> list[dict]:
    return [
        {
            "time": int(time.time()),
            "interval": 60,
            "tags": [f"mac={gth.address}", f"hostname={socket.gethostname()}", f"alias={gth.alias}"],
            **metric,
        }
        for metric in [
            {"name": f"govee.temperature.celsius", "value": gth.temp_celsius},
            {"name": f"govee.humidity.percent", "value": gth.humidity_percent},
            {"name": f"govee.battery.percent", "value": gth.battery_percent},
            {"name": f"govee.rssi", "value": gth.rssi},
        ]
    ]


class Graphite:
    url: str
    user: str | None
    password: str | None
    _session: Soup.Session

    def __init__(self, url: str, user: str | None = None, password: str | None = None):
        self.url = url
        self.user = user
        self.password = password
        self._session = Soup.Session()

        if log.getEffectiveLevel() == logging.DEBUG:
            logger = Soup.Logger.new(Soup.LoggerLogLevel.BODY)
            self._session.add_feature(logger)

    async def send_message(self, gth: Gth):
        uri = GLib.Uri.parse(self.url, GLib.UriFlags.NONE)
        message = Soup.Message.new_from_uri("POST", uri)
        if self.user and self.password:
            assert self._session
            auth_manager = self._session.get_feature(Soup.AuthManager)
            assert auth_manager
            auth = Soup.Auth.new(Soup.AuthBasic, message, "Basic")
            assert auth
            auth.authenticate(self.user, self.password)
            auth_manager.use_auth(message.get_uri(), auth)  # type: ignore

        assert message

        body = json.dumps(to_graphite(gth))

        message.set_request_body_from_bytes("application/json", GLib.Bytes.new(body.encode()))  # type: ignore

        resp_body = await self._session.send_and_read_async(message, GLib.PRIORITY_DEFAULT)  # type: ignore

        if (status := message.get_status()) != Soup.Status.OK:
            log.warning(f"Error Posting to '{self.url}': {Soup.Status.get_phrase(status)} -> {resp_body.get_data().decode()}")
        elif published := json.loads(resp_body.get_data().decode()).get("published"):  # type: ignore
            log.info(f"Published {published} Metric")
